// Autor SERBAN Alexandru-George 335CBb

#include <MD_Parola.h>
#include <MD_MAX72xx.h>
#include <SPI.h>
#include <SD.h>
#include <Wire.h>
#include <DS3231.h>

#define HARDWARE_TYPE MD_MAX72XX::FC16_HW		// tipul matricilor LED folosite
#define MAX_DEVICES 8					// numarul de matrici LED

#define CS_PIN    49					// pinul de la Arduino folosit pentru CS la matricile LED in interfata SPI 
#define NORMAL_SPEED_TEXT "30"				// viteza default la deplasare text stanga
#define LAST_CMD " "					

MD_Parola P = MD_Parola(HARDWARE_TYPE, CS_PIN, MAX_DEVICES);	// P instantiere MD_Parola 

String cmd;

String speedText;
String light_intensityText;

String last_strDisplay;
bool time_cmd;

DS3231 RTC;
bool h12, pm;
bool century;


// initializari
void setup(void) {					 
  cmd = "#0";

  speedText = NORMAL_SPEED_TEXT;
  light_intensityText = "10";

  last_strDisplay = LAST_CMD;
  time_cmd = true;

  pinMode(53, OUTPUT);					// pinul de la Arduino folosit pentru CS la modulul SD card in interfata SPI 

  Serial.begin(9600);					// initializare comunicatie UART pentru serial monitor
  Serial1.begin(9600);					// initializare comunicatie UART pentru bluetooth
  P.begin();						// initializare panou LED-uri

  Wire.begin();						// initializare interfata I2C
}


// functie pentru aflarea token-ului cu indice noToken obtinut dupa split-uirea stringului strToSplit
// delimitat de separatorul sep
String splitStr(String strToSplit, char sep, int noToken) {
  String token = "";
  int currToken = 0;

  for (unsigned int i = 0; i < strlen(*((const char **) &strToSplit)); i++) {
        if (strToSplit[i] == sep) {
          currToken += 1;

          if (currToken > noToken)
            return token;
        }
        else if (currToken == noToken)
          token.concat(strToSplit[i]);
  }

  return token;
}


// functie pentru copierea continutului unui string in alt string
String copyStr(String strToCopy) {
  String strFromCopy = "";

  for (unsigned int i = 0; i < strlen(*((const char**) &strToCopy)); i++)
    strFromCopy.concat(strToCopy[i]);

  return strFromCopy;
}


// functie pentru stergerea caracterelor NL si CR de la finalul unui string 
void strRemoveNLCR(String &strNLCR) {
  if (strNLCR[strNLCR.length() - 2] == '\n' ||
      strNLCR[strNLCR.length() - 2] == '\r')
      strNLCR.remove(strNLCR.length() - 2);

  if (strNLCR[strNLCR.length() - 1] == '\n' ||
      strNLCR[strNLCR.length() - 1] == '\r')
      strNLCR.remove(strNLCR.length() - 1);

  strNLCR.c_str();
}


// functie care verifica daca un string este de tip numeric (contine doar cifre)
bool isNumStr(String strNum) {
  for (unsigned int i = 0; i < strlen(*((const char**) &strNum)); i++) {
    if (!isDigit(strNum[i]))
      return false;
  }

  return true;
}


// functie care afla tokenii in urma split-uirii unui string ce reprezinta comanda data
// (delimitat de spatiu) 
int readCmdStr(String &strCmdValue, String strDisplay) {
  int i = 1;
  while ((strCmdValue = splitStr(strDisplay, ' ', i)) == "")
    i += 1;

  return i;
}


// functie care citeste circuitul RTC (timp, data, temperatura) 
void getEnvData(String &strDisplay) {
  strDisplay = " ";
  strDisplay.concat(RTC.getHour(h12, pm));
  strDisplay.concat(':');

  strDisplay.concat(RTC.getMinute());
  strDisplay.concat(':');

  strDisplay.concat(RTC.getSecond());
  strDisplay.concat("  ");

  strDisplay.concat(RTC.getDate());
  strDisplay.concat('.');

  strDisplay.concat(RTC.getMonth(century));
  strDisplay.concat('.');

  strDisplay.concat(RTC.getYear());
  strDisplay.concat("  ");

  strDisplay.concat(" Temp = ");
  strDisplay.concat(RTC.getTemperature());
  strDisplay.concat(" grdC  ");
}


// functie care programeaza RTC (timp si data)
void setEnvData(String &strDisplay) {
      String hour, minute, second;
      String date, month, year;

      strDisplay = strDisplay.substring(readCmdStr(hour, strDisplay) + strlen(*((const char **) &cmd)));
      strDisplay = strDisplay.substring(readCmdStr(minute, strDisplay) + strlen(*((const char**) &hour)));
      strDisplay = strDisplay.substring(readCmdStr(second, strDisplay) + strlen(*((const char**) &minute)));

      strDisplay = strDisplay.substring(readCmdStr(date, strDisplay) + strlen(*((const char**) &second)));
      strDisplay = strDisplay.substring(readCmdStr(month, strDisplay) + strlen(*((const char**) &date)));
      strDisplay = strDisplay.substring(readCmdStr(year, strDisplay) + strlen(*((const char**) &month)));

      RTC.setYear(year.toInt());
      RTC.setMonth(month.toInt());
      RTC.setDate(date.toInt());

      RTC.setHour(hour.toInt());
      RTC.setMinute(minute.toInt());
      RTC.setSecond(second.toInt());
}


// functie care seteaza string-ul cu datele de la RTC actiualizate 
bool updateTimeCommand(String &strDisplay) {
  if (!cmd.compareTo("#0") || !cmd.compareTo("#9") || (!cmd.compareTo("#3") && time_cmd)
      || (!cmd.compareTo("#4") && time_cmd) || (!cmd.compareTo("#6") && time_cmd)
      || (!cmd.compareTo("#7") && time_cmd) || (!cmd.compareTo("#8") && time_cmd)
      || (!cmd.compareTo("#10") && time_cmd)) {

    getEnvData(strDisplay);
    return true;
  }

  return false;
}


// functie care verifica daca SD card a fost initializat
bool checkSD() {
  if (!SD.begin(53)) {
    Serial.println("Could not initialize SD card");
    Serial1.println("Could not initialize SD card");
    return false;
  }

  return true;
}


// functie care citeste fisier din SD card
void readSD(String &strDisplay, String fileName_SD) {
  if (!checkSD())
    return;

  File file_SD;
  file_SD = SD.open(fileName_SD);

  if (!file_SD) {
    strDisplay = "File ";
    strDisplay.concat(fileName_SD);
    strDisplay.concat(" does not exist on SD card");

    return;
  }

  strDisplay = "";
  while (file_SD.available()) {
    strDisplay.concat((char) file_SD.read());

    if (strDisplay[strDisplay.length() - 1] == '\r'
        || strDisplay[strDisplay.length() - 1] == '\n')
      strDisplay.remove(strDisplay.length() - 1);
  }

  file_SD.close();
}


// functie care scrie in fisierul din SD card
void writeSD(String &strDisplay, String writeStr_SD, String fileName_SD) {
  if (!checkSD())
    return;

  File file_SD;
  file_SD = SD.open(fileName_SD, FILE_WRITE);

  if (!file_SD) {
    strDisplay = "Couldn't create write file ";
    strDisplay.concat(fileName_SD);

    return;
  }

  file_SD.print(writeStr_SD.c_str());
  file_SD.close();
}


// functie care sterge fisier din SD card
void removeFileSD(String fileName_SD) {
  if (!checkSD())
    return;

  if(!SD.remove(fileName_SD)) {
    Serial.println("Could not remove file");
    Serial1.println("Could not remove file");
  }
}


// functie care listeaza fisierele si directoarele din radacina SD card
// la serial monitor si la bluetooth
void listDirSD(File dir) {
  File currFile = dir.openNextFile();

  if (!currFile)
    return;

  Serial.println(currFile.name());
  Serial1.println(currFile.name());
  listDirSD(dir);
}


// functie care determina tipul comenzii primite si retine rezultatul acesteia in string-ul strDisplay
void parseCommand(String &strDisplay) {
    cmd = splitStr(strDisplay, ' ', 0);

    if (!cmd.compareTo("#0")) {
        getEnvData(strDisplay);
    }

    else if (!cmd.compareTo("#1")) {
        strDisplay = strDisplay.substring(strlen(*((const char **) &cmd)));
    }

    else if (!cmd.compareTo("#2")) {
        int i = readCmdStr(speedText, strDisplay);
        strDisplay = strDisplay.substring(i + strlen(*((const char **) &cmd)) + strlen(*((const char **) &speedText)));
    }

    else if (!cmd.compareTo("#3")) {
        readCmdStr(speedText, strDisplay);
        strDisplay = LAST_CMD;
    }

    else if (!cmd.compareTo("#4")) {
        readCmdStr(light_intensityText, strDisplay);

        if (isNumStr(light_intensityText)) {
          P.setIntensity(light_intensityText.toInt());
          strDisplay = LAST_CMD;
        }
        else
          strDisplay = "Text light intensity must be a number";
    }

    else if (!cmd.compareTo("#5")) {
      String fileName_SD;
      readCmdStr(fileName_SD, strDisplay);
      readSD(strDisplay, fileName_SD);
    }

    else if (!cmd.compareTo("#6")) {
      String fileName_SD;
      String writeStr_SD;

      int i = readCmdStr(fileName_SD, strDisplay);
      writeStr_SD = strDisplay.substring(i + strlen(*((const char **) &cmd)) + strlen(*((const char **) &fileName_SD)));

      strDisplay = LAST_CMD;
      writeSD(strDisplay, writeStr_SD, fileName_SD);
    }

    else if (!cmd.compareTo("#7")) {
      String fileName_SD;
      readCmdStr(fileName_SD, strDisplay);

      strDisplay = LAST_CMD;
      removeFileSD(fileName_SD);
    }

    else if (!cmd.compareTo("#8")) {
      strDisplay = LAST_CMD;
      if (!checkSD())
        return;

      listDirSD(SD.open("/"));
    }

    else if (!cmd.compareTo("#9")) {      
      setEnvData(strDisplay);
      strDisplay = LAST_CMD;
    }

    else if (!cmd.compareTo("#10")) {
      P.setInvert(!P.getInvert());
      strDisplay = LAST_CMD;
    }

    else
        strDisplay = "Wrong command!";
}


// functia principala in care se asteapta comanda de executat
// fie de la serial monitor sau de la bluetooth 
void loop(void) {
  String strDisplay = LAST_CMD;

  while (Serial.available()) {
    strDisplay = Serial.readString();
    strRemoveNLCR(strDisplay);
    parseCommand(strDisplay);
  }

  while (Serial1.available()) {
    strDisplay = Serial1.readString();
    strRemoveNLCR(strDisplay);
    parseCommand(strDisplay);
  }

  time_cmd = updateTimeCommand(strDisplay);
  if (strDisplay != LAST_CMD)
    last_strDisplay = copyStr(strDisplay);

  if (P.displayAnimate()) {
    P.displayText(*((const char**) &last_strDisplay), PA_LEFT, speedText.toInt(), 0, PA_SCROLL_LEFT);
    P.displayClear();
  }

}
